This question requires a numerical response in the form of an integer or decimal (e.g., -12, 0, 1, 2, 3.45, ...).
Respond with just your number on a single line.
If your response is equivalent to zero, report '0'

{% if include_comment %}
After the answer, put a comment explaining your choice on the next line.
{% endif %}