{% for index, interview in interviews.items() %}
    {% if interview.exceptions != {} %}
       <div class="interview">Interview: {{ index }} </div>
       Model: {{ interview.model.model }}
    <h1>Failing questions</h1>
    {% endif %}
    {% for question, exceptions in interview.exceptions.items() %}
        {% include 'interview_details.html' %}
    {% endfor %}            
{% endfor %}